#include "Line.h"

#define _USE_MATH_DEFINES
#include <cmath>

namespace mas {

Line::Line ():
  a_ (0.0),
  b_ (1.0),
  c_ (0.0)
{
}

Line::Line (const Point& p1, const Point& p2):
  a_ (p2.Y () - p1.Y ()),
  b_ (p1.X () - p2.X ()),
  c_ (p2.X ()*p1.Y () - p2.Y () * p1.X ())
{
}

Line::Line (const Point& p, double angle):
  a_ (::sin (angle)),
  b_ (-::cos (angle)),
  c_ (p.Y () * -b_ - p.X () * a_)  // Attention: depends on initialization order!
{
}

Line::Line (const Line& other):
  a_ (other.a_),
  b_ (other.b_),
  c_ (other.c_)
{
}

Line::Line (double a, double b, double c):
  a_ (a),
  b_ (b),
  c_ (c)
{
}

Line::~Line ()
{
}

Line* Line::Clone () const
{
  return new Line (*this);
}

bool Line::operator == (const Line &other) const
{
  return a_*other.b_ == b_*other.a_ &&
             a_*other.c_ == c_*other.a_ &&
             b_*other.c_ == c_*other.b_;             
}

bool Line::operator != (const Line &other) const
{
  return ! operator == (other);            
}

double Line::Angle () const
{
  return b_ == 0.0 ? M_PI_2 : ::atan (-a_/b_);
}

double Line::X0 () const
{
  return -c_ / a_;
}

double Line::Y0 () const
{
  return -c_ / b_;
}

bool Line::Vertical () const
{
  return b_ == 0.0;
}

bool Line::Horizontal () const
{
  return a_ == 0.0;
}

bool Line::Has (const Point& p) const
{
  return Value (p) == 0.0;
}

bool Line::Between (const Point& p1, const Point& p2) const
{
  return Value(p1) * Value(p2) <= 0.0;
}

bool Line::SameSide (const Point& p1, const Point& p2) const
{
  return Value(p1) * Value(p2) >= 0.0;
}

bool Line::IsParallel (const Line& other) const
{
  return a_*other.b_ == b_*other.a_; 
}

Point Line::Intersection (const Line& other) const
{
  double x = (b_ * other.c_ - c_ * other.b_) / (a_ * other.b_ - b_ * other.a_);
  double y = (a_ * other.c_ - c_ * other.a_) / (b_ * other.a_ - a_ * other.b_);
  return Point (x, y);
}

Line Line::Parallel (const Point& p) const
{
  return Line (a_, b_, -(a_ * p.X () + b_ * p.Y ()));
}

Line Line::Perpendicular (const Point& p) const
{
  return Line (-b_, a_, -(-b_ * p.X() + a_ * p.Y()));
}

double Line::DistanceTo (const Point& p) const
{
  return Intersection(Perpendicular(p)).DistanceTo(p);
}

void Line::Scale (double fx, double fy)
{
  a_ *= fy;
  b_ *= fx;
  c_ *= fx * fy;
}

void Line::Translate (double dx, double dy)
{
  c_ -= a_ * dx + b_* dy;
}

void Line::Rotate (double angle)
{
  double a0 = a_;
  double b0 = b_;
  a_ = a0 * ::cos(angle) - b0 * ::sin(angle);
  b_ = a0 * ::sin(angle) + b0 * ::cos(angle);
}

void Line::Write (std::ostream& output) const
{
  output << a_ << " " << b_ << " " << c_ ;
}

double Line::Value (const Point& p) const
{
  return a_ * p.X() + b_ * p.Y() + c_;
}

void Line::Draw (Canvas& x) const
{
  x.DrawLine (*this);
}

}
